//////////////////////////////////////////////////////////////////////////
//	HotspotFisher.cpp
//	Description:
//		Identify recombination hotspots, using penalized truncated pairwise recombination rate.
//	Author:
//		Jun Li, jun-li00@mails.tsinghua.edu.cn
//		Bioinformatics Division, TNLIST and Department of Automation, Tsinghua University
//		Beijing 100084, P. R. China
//	Important Note:
//		Part of this code (mainly 'CalHapLikMis' and 'CalHapLikMis') is copied (with slight modification) from McVean's LDhat2.
//		Thanks to him.
//	Verson:
//		1.0		Folder version & Applied to either haploid or diploid data
//				Be able to identify single or multiple hotspots.
//		Difference from Version 0.0:
//		Output delta likelihood vector, and do not output stat_popu*.
//	Last Modified:
//		March 3, 2006
//	Usage:
//		type "HotspotFisher" in the command line to get the usage.
//////////////////////////////////////////////////////////////////////////

//------------------------------ includes -------------------------------------
#include <iostream.h>
#include <iomanip.h>
#include <fstream.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "readfile.h"
#include "fisher.h"
#include "vectool.h"
#include "bootstrap.h"
#include "singlelik.h"
#include "linu.h"
#include "regliktool.h"
#include "searchtool.h"
#include "truncliktool.h"
#include "mixliktool.h"

//--------------------------- declarations ------------------------------------
void HelpInfo(void);
void ProcFile(LikTable& lt, RateSearch& rs, FileNames& fn, BootPara& bp, WinWeight& ww, OptSr& os, double MAF);
void SetWt(WinWeight& ww);

//------------------------------ main function --------------------------------
int main(int argc, char *argv[])
{
	//////////////////////////////////////////////////////////////////////////
	// justify the arguments
	if (argc != 5 && argc != 7 && argc != 9 && argc != 11 && argc != 13)
	{
		HelpInfo();
		exit(0);
	}

	//////////////////////////////////////////////////////////////////////////
	// convert the arguments
	WinWeight ww;
	FileNames fn;
	strcpy(fn.likFileName, argv[1]);
	strcpy(fn.locsFileName, argv[2]);
	strcpy(fn.seqFileName, argv[3]);
	strcpy(fn.outFileName, argv[4]);

	//////// default values
	double MAF = 0.05;
	ww.len = 7;
	ww.nsperc = 0.1;
	ww.compLen = 0.1;
	double thrd = 26.0;
	int popuSize = 10000;

	for (int pi = 5; pi < argc; pi += 2)
	{
		if (strcmp(argv[pi], "-F") == 0)
		{
			MAF = atof(argv[pi+1]);
			if (MAF >= 0.5)
			{
				cerr << "MAF should be less than 0.5!" << endl;
				HelpInfo();
				exit(0);
			}
		}
		else if (strcmp(argv[pi], "-W") == 0)
		{
			ww.len = atoi(argv[pi+1]);
			if (ww.len <= 0)
			{
				cerr << "Window length should be no positive integer!" << endl;
				HelpInfo();
				exit(0);
			}
		}
		else if (strcmp(argv[pi], "-T") == 0)
		{
			thrd = atof(argv[pi+1]);
			if (thrd <= 0.0)
			{
				cerr << "LLR threshold should be positive!" << endl;
				HelpInfo();
				exit(0);
			}
		}
		else if (strcmp(argv[pi], "-P") == 0)
		{
			popuSize = atoi(argv[pi+1]);
			if (popuSize <= 0)
			{
				cerr << "Effective population size should be a positive integer!" << endl;
				HelpInfo();
				exit(0);
			}
		}
		else
		{
			HelpInfo();
			exit(0);
		}
	}

	ww.grid = 0.20;
	ww.minHotLen = 4;
	ww.maxHotLen = 12;

	//--------------------- this parameters are set when debug -------------------
	//////////////////////////////////////////////////////////////////////////
	// set search parameters
	OptSr os;
	os.dur = 5;
	os.bkSt = 130;
	os.xdur = 5;
	os.maxIter = 100;
	os.ngap = 4;
	os.higher = 15.0;
	os.perc = 0.15;
	os.bkthrd = 4.0;
	os.amp = (popuSize + 0.0) / 10000.0;
	os.hotperc = 0.50;
//	os.likLenAdd = 0.01;
	
	//////////////////////////////////////////////////////////////////////////
	// set bootstrap paramters
	BootPara bp;
	bp.bootType = 0;	// no bootstrap
	bp.numBoot = 1;
	bp.randSeed = 13427;
	srand(bp.randSeed);

	//////////////////////////////////////////////////////////////////////////
	// set pairwise likelihood parameters
	ww.thrd = thrd;
	ww.higher = 10.0;

	ww.wt = new double[ww.len];
	SetWt(ww);
	
	ww.ewt = new double[ww.len-1];
	EdgeWt(ww);
	if (DEBUG != 2)
	{
		cout << "Edge correction parameters are:";
		for (int wi = 1; wi <= ww.len-1; wi++)
		{
			cout << "\t" << ww.ewt[wi-1];
		}
		cout << endl;
	}

	//////////////////////////////////////////////////////////////////////////
	// the recombination rate that we will search
	RateSearch rs;
	rs.nRate = 200;
	rs.maxRate = 150.0;
	rs.minRate = 0.015;
	rs.rate = new double[rs.nRate];		// calculate the recombination rate that will be searched
	RateVect(rs);
	//---------------------- end of set parameter -----------------------------

	//////////////////////////////////////////////////////////////////////////
	// read likelihood file
	LikTable lt;
	ifstream likFile(fn.likFileName);
//	if (likFile.fail())
//	{
//		cerr << "Error opening file " << fn.likFileName << ".!" << endl;
//		exit(0);
//	}

	ReadBasicPara(likFile, lt);	// read in the basic parameters
	
	lt.recombRate = new double[lt.numRate];
	ReadRecombRate(likFile, lt);	// read in the recombination rate
	RetouchRecombRate(lt);

	lt.comb = new int[4 * lt.numComb];
	lt.lik = new double[lt.numRate * lt.numComb];
	ReadLik(likFile, lt);	// read in the combinations and the likelihoods

	//////////////////////////////////////////////////////////////////////////
	// proceeding the file
	ProcFile(lt, rs, fn, bp, ww, os, MAF);

	//////////////////////////////////////////////////////////////////////////
	// free space
	delete[] ww.wt;
	delete[] ww.ewt;
	delete[] lt.recombRate;
	delete[] lt.comb;
	delete[] lt.lik;
	delete[] rs.rate;

	if (DEBUG != 2)
	{
		cerr << "Program done successfully." << endl;
	}
	else
	{
		cout << "Program done successfully." << endl;
	}
	return 0;
}


//////////////////////////////////////////////////////////////////////////
// output usage information
void HelpInfo(void)
{
	cerr << "HotspotFisher" << endl
		 << "Description:"<< endl
		 << "  Detect recombination hotspots, using TWPLL." << endl
		 << "Author:" << endl
		 << "  Li Jun, jun-li00@mails.tsinghua.edu.cn" << endl
		 << "  Bioinformatics Division, TNLIST and Department of Automation" << endl
		 << "  Tsinghua University, Beijing 100084, P. R. China" << endl
		 << "Version:" << endl
		 << "  1.3" << endl
		 << "Last Modified:" << endl
		 << "  May 29, 2006" << endl
		 << endl
		 << "Usage:" << endl
		 << "% ./HotspotFisher lookupTable inLocs inData outRes effWinLen LLRThrd [-F MAF] [-W windLen] [-T LLRthrd] [-P effPopuSize]" << endl
		 << "The parameters in the brackets are optinal." << endl
		 << endl
		 << "e.g." << endl
		 << "  HotspotFisher lk_n180_t0.001 locs hap res" << endl
		 << endl
		 << "e.g." << endl
		 << "  HotspotFisher lk_n180_t0.001 locs hap res -F 0.10 -P 12000" << endl
		 << endl
		 << "Please note that the format of input files are the same as LDhat 2.0." << endl;
}

//////////////////////////////////////////////////////////////////////////
// Proceed the calculation and output of one file
void ProcFile(LikTable& lt, RateSearch& rs, FileNames& fn, BootPara& bp, WinWeight& ww, OptSr& os, double MAF)
{
	//////////////////////////////////////////////////////////////////////////
	// read genotype file
	Sequ osq;
	ifstream seqFile(fn.seqFileName);
//	if (seqFile.fail())
//	{
//		cout << "Error opening file " << fn.seqFileName << ".!" << endl;
//		exit(0);
//	}

	ReadSeqPara(seqFile, osq);

	osq.seq = new int[osq.numSamp * osq.numSNP]; 
	ReadSeq(seqFile, osq);

	//////////////////////////////////////////////////////////////////////////
	// read SNP locs file
	Loci olc;
	ifstream locsFile(fn.locsFileName);
//	if (locsFile.fail())
//	{
//		cout << "Error opening file " << fn.locsFileName << ".!" << endl;
//		exit(0);
//	}

	ReadLocsPara(locsFile, olc);
	if (olc.numSNP != osq.numSNP)
	{
		cerr << "Number of SNP in the sequence file does not correspond to that in the locs files!" << endl;
		exit(0);
	}
	olc.locs = new double[olc.numSNP];
	ReadLocs(locsFile, olc);

	//////////////////////////////////////////////////////////////////////////
	// select SNPs
	IntVect iv;
	iv.num = osq.numSNP;
	iv.vect = new int[iv.num];

	Sequ sq;
	sq.numSamp = osq.numSamp;
	sq.hg = osq.hg;
	sq.numSNP = JudgeSites(iv, osq, olc, MAF);
	sq.seq = new int[sq.numSamp * sq.numSNP];
	SelSNPSeq(sq, osq, iv);

	Loci lc;
	lc.numSNP = sq.numSNP;
	lc.locs = new double[lc.numSNP];
	SelSNPLocs(lc, olc, iv);

	//////////////////////////////////////////////////////////////////////////
	// free space
	delete[] osq.seq;
	delete[] olc.locs;
	delete[] iv.vect;

	//////////////////////////////////////////////////////////////////////////
	// allocation
	// number of types of frequencies
	int numFreq = 9;
	if (sq.hg == 2)
	{
		numFreq = 16;
	}
	int *freq = new int[numFreq];	// combination of 2 SNPs
	DataLikMatrix lm;
	lm.numSNP = sq.numSNP;
	lm.numRate = lt.numRate;
	lm.typInd = new int[lm.numSNP * (lm.numSNP-1) / 2];	// genotype indice
	lm.genoInd = new int[lm.numSNP * (lm.numSNP-1) / 2 * numFreq];	// genotype kinds
	lm.likMatrix = new double[lm.numSNP * (lm.numSNP-1) / 2 * lm.numRate];	// likelihood matrix

	//////////////////////////////////////////////////////////////////////////
	// begin bootstrap
	double *lnfac = new double[lt.numHap + 2];
	GetLnfac(lnfac, lt.numHap + 2);
	int *ind = new int[sq.numSamp];
	double *lknew = new double[lt.numRate];	// final likelihoods for the use of CalDipLik
	double *lkres = new double[lt.numRate];	// temp likelihoods
	int *hetInd = new int[sq.numSNP];
	int *homoInd = new int[sq.numSNP];
	
	int iBoot;
	for (iBoot = 0; iBoot < bp.numBoot; iBoot++)
	{
		//////////////////////////////////////////////////////////////////////////
		// get bootstrap indice
		if (bp.bootType == 1)
		{
			cout << endl << endl << "Bootstrap " << iBoot << "..." << endl;
		}
		else if (bp.bootType == 2)
		{
			cout << endl << endl << "Subsample " << iBoot << "..." << endl;
		}
		else
		{
			cout << endl << endl << "Original data." << endl;
		}
		Bootstrap(ind, sq.numSamp, bp.bootType);	// get the bootstrap indices

		int numHet = SnpHetero(hetInd, homoInd, sq.seq, sq.numSNP, ind, int(lt.numHap/sq.hg), ALLELE_FREQ_THRD);	// get the hetero SNP indices
		if (numHet != sq.numSNP)
		{
			cerr << "DISCARD ONE BOOTSTRAP!!" << endl;
			iBoot--;
			continue;
		}

		//////////////////////////////////////////////////////////////////////////
		// calculating likelihoods
		cout << "Calculating likelihood... " << endl;
		int procI = 0;
		int iS, jS;
		for (iS = 0; iS < sq.numSNP - 1; iS++)
		{
			for (jS = iS + 1; jS < sq.numSNP; jS++)
			{
				if (DEBUG != 2)
				{
					cout << ".";
				}
				if (DEBUG == 1)
				{
					cout << endl << "SNP " << iS << " & " << jS << ":";
				}
				// calculate likelihoods
				if (sq.hg == 2)
				{
					GetPairFreqGeno(freq, sq.seq, sq.numSNP, ind, lt.numHap/2, iS, jS);	// get genotype pair frequencies
					CalDipLik(lknew, lkres, freq, lnfac, lt.numHap/2, lt.numRate, lt.lik);	// calculate the genotype likelihoods (perhaps) with missing data
				}
				else
				{
					GetPairFreqHap(freq, sq.seq, sq.numSNP, ind, lt.numHap, iS, jS);	// get haplotype pair frequencies
					CalHapLik(lknew, freq, lnfac, lt.numHap, lt.numRate, lt.lik);	// calculate the haplotype likelihoods
				}
				
				CopyVect(lm.genoInd + procI * numFreq, freq, numFreq);		// genotype pair
				lm.typInd[angInd2(iS, jS, sq.numSNP)] = procI;	// genotype pair index
				CopyVect(lm.likMatrix + procI * lt.numRate, lknew, lt.numRate);	// likelihood

				procI++;
			}
			if (DEBUG != 2)
			{
				cout << endl;
			}
		}
		if (DEBUG != 2)
		{
			cerr << "OK!" << endl;
		}
		else
		{
			cout << "OK!" << endl;
		}
		if (DEBUG == 1)
		{
			cerr << endl;
			OutLikFile(lm, procI, lt.numRate, sq.hg);
		}

		//////////////////////////////////////////////////////////////////////////
		// identify hotspots
		HotReg hr;
		MixHot(rs, lc, lm, lt, ww, hr, os, fn);
	}

	//////////////////////////////////////////////////////////////////////////
	// free space
	delete[] sq.seq;
	delete[] lc.locs;
	delete[] freq;
	delete[] lm.typInd;
	delete[] lm.genoInd;
	delete[] lm.likMatrix;
	delete[] lnfac;
	delete[] ind;
	delete[] lknew;
	delete[] lkres;
	delete[] hetInd;
	delete[] homoInd;
}

//////////////////////////////////////////////////////////////////////////
// set weight
void SetWt(WinWeight& ww)
{
	int wi;
	for (wi = 0; wi < ww.len; wi++)
	{
		//ww.wt[wi] = (ww.len - wi + 0.0) / ww.len;
		ww.wt[wi] = 1.0 / (wi + 1.0);
	}
}
